package de.uni_passau.fim.seibt.gitwrapper.process;

import java.io.File;
import java.io.IOException;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.io.IOUtils;

import static de.uni_passau.fim.seibt.gitwrapper.repo.GitWrapper.EXIT_FAIL;
import static java.nio.charset.StandardCharsets.UTF_8;
import static java.util.logging.Level.WARNING;

/**
 * Contains static methods for executing commands and gathering their output and exit code.
 */
public class ProcessExecutor {

    private static final Logger LOG = Logger.getLogger(ProcessExecutor.class.getCanonicalName());
    private static final ExecutorService exService = Executors.newCachedThreadPool();

    /**
     * The result of a process execution.
     */
    public static final class ExecRes {
        public final int exitCode;
        public final String stdOut;
        public final String stdErr;

        private ExecRes(int exitCode, String stdOut, String stdErr) {
            this.exitCode = exitCode;
            this.stdOut = stdOut;
            this.stdErr = stdErr;
        }
    }

    /**
     * Executes the given <code>command</code>. If <code>command</code> is <code>null</code> an empty {@link Optional}
     * will be returned.
     *
     * @param workingDir
     *         the working directory for the execution
     * @param redirectError
     *         whether to merge the standard output and error streams
     * @param command
     *         the command to be executed
     * @return the result of the execution or an empty {@link Optional} if there is an exception executing the command
     */
    public static Optional<ExecRes> exec(File workingDir, boolean redirectError, List<String> command) {

        if (command == null) {
            return Optional.empty();
        }

        ProcessBuilder builder = new ProcessBuilder();

        builder.command(command);
        builder.directory(workingDir);
        builder.redirectErrorStream(redirectError);

        return exec(builder);
    }

    /**
     * Executes the command represented by the given <code>builder</code>.
     *
     * @param builder
     *         the {@link ProcessBuilder} to be executed
     * @return the result of the execution or an empty {@link Optional} if there is an exception executing the command
     */
    public static Optional<ExecRes> exec(ProcessBuilder builder) {
        String cmd = String.join(" ", builder.command());
        LOG.fine(() -> String.format("Executing '%s' in '%s'.", cmd, builder.directory().getAbsolutePath()));

        try {
            Process p = builder.start();

            String stdOut;
            String stdErr;

            if (builder.redirectErrorStream()) {
                stdOut = IOUtils.toString(p.getInputStream(), UTF_8).trim();
                stdErr = IOUtils.toString(p.getErrorStream(), UTF_8).trim();
            } else {
                Future<String> outF = exService.submit(() -> IOUtils.toString(p.getInputStream(), UTF_8).trim());
                Future<String> errF = exService.submit(() -> IOUtils.toString(p.getErrorStream(), UTF_8).trim());

                try {
                    stdOut = outF.get();
                    stdErr = errF.get();
                } catch (InterruptedException e) {
                    LOG.log(WARNING, e, () -> String.format("Interrupted while waiting for '%s' to finish.", cmd));

                    outF.cancel(true);
                    errF.cancel(true);
                    p.destroyForcibly();
                    return Optional.empty();
                } catch (ExecutionException e) {
                    LOG.log(WARNING, e, () -> String.format("Exception while reading the output of '%s'.", cmd));

                    outF.cancel(true);
                    errF.cancel(true);
                    p.destroyForcibly();
                    return Optional.empty();
                }
            }

            IOUtils.closeQuietly(p.getInputStream());
            IOUtils.closeQuietly(p.getErrorStream());
            IOUtils.closeQuietly(p.getOutputStream());

            int exitCode;
            
            try {
                exitCode = p.waitFor();
            } catch (InterruptedException e) {
                LOG.log(WARNING, e, () -> String.format("Interrupted while waiting for '%s' to finish.", cmd));

                p.destroyForcibly();
                return Optional.empty();
            }

            ExecRes res = new ExecRes(exitCode, stdOut, stdErr);

            LOG.fine(() -> String.format("Execution of '%s' returned exit code %d.", cmd, res.exitCode));

            if (res.stdOut.isEmpty()) {
                LOG.finer(() -> String.format("Execution of '%s' returned no standard output.", cmd));
            } else {
                LOG.finer(() -> String.format("Execution of '%s' returned standard output.", cmd));
                LOG.finest(() -> String.format("Standard output of '%s' was:%n%s", cmd, res.stdOut));
            }

            if (res.stdErr.isEmpty()) {
                LOG.finer(() -> String.format("Execution of '%s' returned no standard error output.", cmd));
            } else {
                LOG.finer(() -> String.format("Execution of '%s' returned standard error output.", cmd));
                LOG.finest(() -> String.format("Standard output of '%s' was:%n%s", cmd, res.stdErr));
            }

            return Optional.of(res);
        } catch (IOException e) {
            LOG.log(Level.SEVERE, e, () -> String.format("Exception executing '%s'.", cmd));
            return Optional.empty();
        }
    }

    /**
     * Returns a modifiable {@link LinkedList} containing the given elements.
     *
     * @param ts
     *         the elements to return in a {@link LinkedList}
     * @param <T>
     *         the type of the elements
     * @return a {@link LinkedList} containing the elements
     */
    @SafeVarargs
    public static <T> LinkedList<T> asList(T... ts) {
        LinkedList<T> l = new LinkedList<>(); // we expect things to be inserted in front and back

        if (ts == null) {
            return l;
        }

        Collections.addAll(l, ts);

        return l;
    }
}
